package org.rapidoid.app;

/*
 * #%L
 * rapidoid-app
 * %%
 * Copyright (C) 2014 - 2015 Nikolche Mihajlovski and contributors
 * %%
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * #L%
 */

import java.util.Map;
import java.util.Map.Entry;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.script.CompiledScript;
import javax.script.ScriptException;
import javax.script.SimpleBindings;

import org.rapidoid.annotation.Authors;
import org.rapidoid.annotation.Since;
import org.rapidoid.beany.BeanProperties;
import org.rapidoid.beany.Beany;
import org.rapidoid.beany.Prop;
import org.rapidoid.config.Conf;
import org.rapidoid.dispatch.DispatchResult;
import org.rapidoid.dispatch.PojoDispatchException;
import org.rapidoid.dispatch.PojoDispatcher;
import org.rapidoid.dispatch.PojoHandlerNotFoundException;
import org.rapidoid.dispatch.PojoRequest;
import org.rapidoid.dispatch.impl.DispatchReqKind;
import org.rapidoid.http.Handler;
import org.rapidoid.http.HttpExchange;
import org.rapidoid.http.HttpExchangeImpl;
import org.rapidoid.http.HttpExchangeInternals;
import org.rapidoid.http.HttpNotFoundException;
import org.rapidoid.io.CustomizableClassLoader;
import org.rapidoid.io.Res;
import org.rapidoid.jackson.JSON;
import org.rapidoid.log.Log;
import org.rapidoid.plugins.templates.Templates;
import org.rapidoid.util.Constants;
import org.rapidoid.util.U;
import org.rapidoid.util.UTILS;
import org.rapidoid.webapp.AppCtx;
import org.rapidoid.webapp.WebApp;
import org.rapidoid.webapp.WebEventReq;
import org.rapidoid.webapp.WebReq;

@Authors("Nikolche Mihajlovski")
@Since("2.0.0")
public class AppHandler implements Handler {

	private static final String PAGE_RELOAD = "<h2>&nbsp;Reloading...</h2><script>location.reload();</script>";

	private static final Pattern DIRECTIVE = Pattern.compile("\\s*\\Q<!--#\\E\\s*(\\{.+\\})\\s*\\Q-->\\E\\s*");

	private CustomizableClassLoader classLoader;

	public AppHandler() {
		this(null);
	}

	public AppHandler(CustomizableClassLoader classLoader) {
		this.classLoader = classLoader;
	}

	@Override
	public Object handle(final HttpExchange x) throws Exception {

		HttpExchangeInternals xi = (HttpExchangeInternals) x;
		xi.setClassLoader(classLoader);

		Object result;

		try {
			result = processReq(x);
		} catch (Exception e) {
			if (UTILS.rootCause(e) instanceof HttpNotFoundException) {
				throw U.rte(e);
			} else {
				// Log.error("Exception occured while processing request!", UTILS.rootCause(e));
				throw U.rte(e);
			}
		}

		return result;
	}

	public Object processReq(HttpExchange x) {
		Object result = dispatch((HttpExchangeImpl) x);

		if (result != null) {
			return result;
		} else {
			throw x.notFound();
		}
	}

	@SuppressWarnings("unchecked")
	public Object dispatch(HttpExchangeImpl x) {

		// static files

		if (x.isGetReq() && x.serveStaticFile()) {
			return x;
		}

		WebApp app = AppCtx.app();
		PojoDispatcher dispatcher = app.getDispatcher();

		boolean hasEvent = false;

		// Prepare GUI state

		x.loadState();

		// if an event was emitted, process it

		if (x.isPostReq()) {
			String event = x.posted("_event", null);
			if (!U.isEmpty(event)) {
				hasEvent = true;

				String evArgs = x.posted("_args", null);
				Object[] args = evArgs != null ? JSON.jacksonParse(evArgs, Object[].class) : Constants.EMPTY_ARRAY;

				String inputstr = x.posted("_inputs");
				U.notNull(inputstr, "inputs");
				Map<String, Object> inputs = JSON.parse(inputstr, Map.class);

				// bind inputs
				for (Entry<String, Object> e : inputs.entrySet()) {
					String inputId = e.getKey();
					Object value = e.getValue();

					x.locals().put(inputId, UTILS.serializable(value));
				}

				DispatchResult dispatchResult = doDispatch(dispatcher, new WebReq(x));

				if (dispatchResult != null) {
					U.must(dispatchResult.getKind() == DispatchReqKind.PAGE);
				}

				// in case of binding or validation errors
				if (x.hasErrors()) {
					x.json();
					return U.map("!errors", x.errors());
				}

				// call the command handler
				DispatchResult dr = on(x, dispatcher, event, args);
				if (dr == null) {
					x.json();
					Log.warn("No event handler was found!", "event", event, "page", x.path());
					return U.map();
				}
			}
		}

		// dispatch REST services or views (as POJO methods)

		DispatchResult dres = doDispatch(dispatcher, new WebReq(x));
		Map<String, Object> config = dres != null ? dres.getConfig() : null;

		Object result = null;
		if (dres != null) {
			result = dres.getResult();

			if (dres.getKind() == DispatchReqKind.SERVICE) {
				return result;
			}
		}

		// serve dynamic pages from a script

		if (result == null) {
			result = runDynamicScript(x);

			if (result == x) {
				result = desc(x);
			} else if (result instanceof Dollar) {
				result = desc((Dollar) result);
			} else if (result instanceof DollarPage) {
				DollarPage page = (DollarPage) result;
				config = page.getConfig();
				result = page.getValue();
			} else if (result != null) {
				return result;
			}
		}

		if (result == null) {
			// try generic app screens
			result = GenericGUI.genericScreen();
		}

		// serve dynamic pages from file templates

		if (serveDynamicPage(x, result, hasEvent, config)) {
			return x;
		}

		if (result != null) {
			return result;
		}

		throw x.notFound();
	}

	private static Object desc(HttpExchangeImpl x) {
		Map<String, Object> desc = U.map();

		desc.put("verb", x.verb());
		desc.put("uri", x.uri());
		desc.put("path", x.path());
		desc.put("home", x.home());
		desc.put("dev", x.isDevMode());

		boolean loggedIn = AppCtx.isLoggedIn();
		desc.put("loggedIn", loggedIn);
		desc.put("user", loggedIn ? AppCtx.user() : null);

		return GUI.multi(GUI.h2("Request details:"), GUI.grid(desc), GUI.h2("Request params:"), GUI.grid(x.data()),
				GUI.h2("Cookies:"), GUI.grid(x.cookies()));
	}

	private static Object desc(Dollar dollar) {
		Map<String, Object> desc = U.map();
		BeanProperties props = Beany.propertiesOf(dollar);

		for (Prop prop : props) {
			Object val = prop.get(dollar);
			desc.put(prop.getName(), val.getClass().getSimpleName());
		}

		return GUI.multi(GUI.h2("The $ properties:"), GUI.grid(desc), GUI.h2("Bindings:"), GUI.grid(dollar.bindings));
	}

	private DispatchResult doDispatch(PojoDispatcher dispatcher, PojoRequest req) {
		try {
			return dispatcher.dispatch(req);
		} catch (PojoHandlerNotFoundException e) {
			// / just ignore, will try to dispatch a page next...
			return null;
		} catch (PojoDispatchException e) {
			throw U.rte("Dispatch error!", e);
		}
	}

	public boolean serveDynamicPage(HttpExchangeImpl x, Object result, boolean hasEvent, Map<String, Object> config) {
		String filename = x.resourceName() + ".html";
		String firstFile = Conf.templatesPath() + "/" + filename;
		String defaultFile = Conf.templatesPathDefault() + "/" + filename;
		Res res = Res.from(filename, true, firstFile, defaultFile);

		Map<String, Object> model = U.cast(U.map("navbar", true, "login", true, "profile", true));

		if (res.exists()) {
			model.putAll(pageModel(result, res));
		} else if (result != null) {
			model.put("result", result);
			model.put("content", result);
		} else {
			return false;
		}

		WebApp app = AppCtx.app();
		model.put("title", app.getTitle());
		model.put("embedded", hasEvent || x.param("embedded", null) != null);

		// the @Page configuration overrides the previous
		if (config != null) {
			model.putAll(config);
		}

		if (hasEvent) {
			serveEventResponse(x, x.renderPageToHTML(model));
		} else {
			x.renderPage(model);
		}

		return true;
	}

	private void serveEventResponse(HttpExchangeImpl x, String html) {
		x.startResponse(200);
		x.json();

		if (x.redirectUrl() != null) {
			x.writeJSON(U.map("_redirect_", x.redirectUrl()));
		} else {
			Map<String, String> sel = U.map("body", html);
			x.writeJSON(U.map("_sel_", sel, "_state_", x.serializeLocals()));
		}
	}

	private static Map<String, Object> pageModel(Object result, Res resource) {
		String template = U.safe(resource.getContent());

		Map<String, Object> model = U.map("result", result);

		String[] contentParts = template.split("\n", 2);
		if (contentParts.length == 2) {
			String line = contentParts[0];

			Matcher m = DIRECTIVE.matcher(line);
			if (m.matches()) {
				Map<String, Object> directives = JSON.parseMap(m.group(1));
				model.putAll(directives);
				template = contentParts[1]; // without the directive
			}
		}

		String content = Templates.fromString(template).render(model, result);
		model.put("content", content); // content without the directive
		return model;
	}

	public DispatchResult on(HttpExchange x, PojoDispatcher dispatcher, String event, Object[] args) {

		Map<String, Object> state = U.cast(x.locals());
		WebEventReq req = new WebEventReq(x.path(), event.toUpperCase(), args, state);

		return doDispatch(dispatcher, req);
	}

	public static final void reload(HttpExchange x) {
		Map<String, String> sel = U.map("body", PAGE_RELOAD);
		x.writeJSON(U.map("_sel_", sel));
	}

	public Object runDynamicScript(HttpExchangeImpl x) {
		String scriptName = x.isGetReq() ? x.resourceName() : x.verb().toUpperCase() + "_" + x.resourceName();
		String filename = scriptName + ".js";
		String firstFile = Conf.dynamicPath() + "/" + filename;
		String defaultFile = Conf.dynamicPathDefault() + "/" + filename;
		Res res = Res.from(filename, true, firstFile, defaultFile);

		if (!res.exists()) {
			return null;
		}

		String js = res.getContent();
		CompiledScript compiled;
		try {
			compiled = U.compileJS(js);
		} catch (ScriptException e) {
			throw U.rte("Script compilation error!", e);
		}

		Map<String, Object> bindings = U.map();

		for (Entry<String, String> e : x.data().entrySet()) {
			bindings.put("$" + e.getKey(), e.getValue());
		}

		Dollar dollar = new Dollar(x, bindings);

		bindings.put("$", dollar);

		try {
			return compiled.eval(new SimpleBindings(bindings));
		} catch (ScriptException e) {
			throw U.rte("Script execution error!", e);
		}
	}

}
